;;; bencode.el --- bencode library for Emacs

;; Copyright (C) 2012  Walter C. Pelissero

;; Author: Walter C. Pelissero <walter@pelissero.de>
;; Keywords: tools, files

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; b-encode/decode library for Emacs parses torrent files and output
;; their contents in human readable way

;;; Code:

(require 'cl)

(defun be-read-delimited-integer (delimiter)
  (let* ((first (char-after))
	 (negative (eq ?- first))
	 (n 0))
    (when negative
      (forward-char))
    (do ((c (char-after) (progn (forward-char)
				(char-after))))
	((eq c delimiter))
      (setq n (+ (* n 10) (- c ?0))))
    (forward-char)
    (if negative (- n) n)))

(defun be-read-integer ()
  (assert (eq ?i (char-after)))
  (forward-char)
  (be-read-delimited-integer ?e))

(defun be-read-string ()
  (let ((len (be-read-delimited-integer ?:)))
    (prog1
	(buffer-substring (point) (+ (point) len))
      (forward-char len))))

(defun be-read-list ()
  (assert (eq ?l (char-after)))
  (forward-char)
  (loop
   until (eq ?e (char-after))
   collect (be-read-value)
   finally (forward-char)))

(defun be-read-dictionary ()
  (assert (eq ?d (char-after)))
  (forward-char)
  (loop
   until (eq ?e (char-after))
   collect (cons (be-read-value) (be-read-value))
   finally (forward-char)))

(defun be-read-value ()
  (let ((type (char-after)))
    (if (and (<= ?0 type)
	     (<= type ?9))
	(be-read-string)
	(ecase type
	  (?i (be-read-integer))
	  (?l (be-read-list))
	  (?d (be-read-dictionary))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun bittorrent-read-file (pathname)
  (with-temp-buffer
    (insert-file-contents-literally pathname)
    (set-buffer-multibyte nil)
    (be-read-dictionary)))

(defun be-get (key alist)
  (cdr (assoc-string key alist)))

(defun bittorrent-print-info (pathname &optional buffer)
  (interactive "f")
  (let* ((buf (or buffer (generate-new-buffer pathname)))
	 (dict (bittorrent-read-file pathname))
	 (info (be-get "info" dict))
	 (created (be-get "created by" dict))
	 (epoch (be-get "creation date" dict))
	 (annl (be-get "announce-list" dict)))
    (with-current-buffer buf
      (insert
       (format "Name: %s\nCreated" (be-get "name" info))
       (if created
	   (format " by %s" created)
	 "")
       (if epoch
	   (destructuring-bind  (sec min hour day month year wd st tz)
	       (decode-time (list (lsh epoch -16) (logand epoch #xff) 0))
	     (declare (ignore wd st tz))
	     (format " on %d/%d/%d %d:%d:%d" day month year hour min sec))
	 "")
       "\n"
       (format "Encoding: %s\n" (be-get "encoding" dict))
       (format "Tracker: %s\n" (be-get "announce" dict)))
      (when annl
	(insert "Announce List:\n")
	(dolist (entry annl)
	  (insert (format "  %s\n" (if (consp entry)
				       (car entry)
				     entry)))))
      (let ((total 0)
	    (prefix (be-get "name" info))
	    (files (be-get "files" info)))
	(if files
	    (progn
	      (insert (format "%d Files:\n" (length files)))
	      (dolist (f files)
		(let ((len (be-get "length" f)))
		  (incf total len)
		  (insert (format "  %10d %s" len (or prefix "")))
		  (dolist (path (be-get "path" f))
		    (insert (format "/%s" path)))
		  (insert "\n"))))
	  (progn
	    (insert (format "File: %s\n" prefix))
	    (setq total (be-get "length" info))))
	(insert (format "Total: %d bytes\n" total))))))


(require 'dired)
(defun dired-view-torrent (&optional arg)
  "Mail selected file(s) as MIME attachment."
  (interactive "P")
  (let ((files (dired-get-marked-files nil arg)))
    (with-output-to-temp-buffer "*Torrent Info*"
      (if files
	  (dolist (file files)
	    (princ file)
	    (princ "\n")
	    (bittorrent-print-info file standard-output)
	    (princ "\n"))
	(bittorrent-print-info (dired-get-file-for-visit) standard-output)))))

(provide 'bencode)
;;; bencode.el ends here
