 ;;; bencode.lisp --- b-encode/decode functions

 ;;; Copyright (C) 2011 by Walter C. Pelissero

 ;;; Author: Walter C. Pelissero <walter@pelissero.de>
 ;;; Project: bencode

#+cmu (ext:file-comment "$Module: bencode.lisp")

;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1
;;; of the License, or (at your option) any later version.
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with this library; if not, write to the Free
;;; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
;;; 02111-1307 USA

(in-package :bencode)

(defvar *unix-lisp-time-difference*
  (encode-universal-time 0 0 0 1 1 1970 0))

(defun universal->unix-time (time)
  (- time *unix-lisp-time-difference*))

(defun unix->universal-time (time)
  (+ time *unix-lisp-time-difference*))

(defun read-delimited-integer (stream delimiter)
  (loop
     with negative = (and (char= #\- (peek-char nil stream))
			  (read-char stream))
     with n = 0
     for c = (read-char stream)
     until (char= c delimiter)
     do (setf n (+ (* n 10) (digit-char-p c)))
     finally (return (if negative (- n) n))))

(defun be-read-integer (stream)
  (assert (char= #\i (read-char stream)))
  (read-delimited-integer stream #\e))

(defun be-read-string (stream)
  (let ((buffer (make-string (read-delimited-integer stream #\:))))
    (read-sequence buffer stream)
    buffer))

(defun be-read-list (stream)
  (assert (char= #\l (read-char stream)))
  (loop
     until (char= #\e (peek-char nil stream))
     collect (be-read-value stream)
     ;; get rid of the #\e
     finally (read-char stream)))

(defun be-read-dictionary (stream)
  (assert (char= #\d (read-char stream)))
  (loop
     with table = (make-hash-table :test 'equalp)
     until (char= #\e (peek-char nil stream))
     do (setf (gethash (be-read-value stream) table)
	      (be-read-value stream))
     ;; get rid of the #\e
     finally (read-char stream)
       (return table)))

(defun be-read-value (stream)
  (let ((type (peek-char nil stream)))
    (if (digit-char-p type)
	(be-read-string stream)
	(ecase type
	  (#\i (be-read-integer stream))
	  (#\l (be-read-list stream))
	  (#\d (be-read-dictionary stream))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro delimited-value ((stream start) &body forms)
  `(progn
     (write-char ,start ,stream)
     ,@forms
     (write-char #\e ,stream)))

(defun be-write-value (stream value)
  (typecase value
    (integer (be-write-integer stream value))
    (string (be-write-string stream value))
    (list (be-write-list stream value))
    (hash-table (be-write-dictionary stream value))))

(defun be-write-integer (stream value)
  (delimited-value (stream #\i)
    (format stream "~D" value)))

(defun be-write-string (stream string)
  (format stream "~D:~A" (length string) string))

(defun be-write-list (stream list)
  (delimited-value (stream #\l)
    (dolist (e list)
      (be-write-value stream e))))

(defun be-write-dictionary (stream dict)
  (delimited-value (stream #\d)
    (mapc #'(lambda (e)
	      (destructuring-bind (key . value) e
		(be-write-value stream key)
		(be-write-value stream value)))
	  ;; dictionaries must appear in lexicographical order
	  (sort (let ((alist '()))
		  (maphash #'(lambda (key value)
			       (push (cons key value) alist))
			   dict)
		  alist)
		#'string< :key #'car))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun bittorrent-read-file (pathname)
  (with-open-file (stream pathname)
    (be-read-dictionary stream)))

(defun bittorrent-write-file (pathname data &key (if-exists :error))
  (assert (hash-table-p data))
  (with-open-file (stream pathname :direction :output :if-exists if-exists)
    (be-write-dictionary stream data))
  (values))

(defun bittorrent-print-info (pathname &optional (stream *standard-output*))
  (let* ((dict (bittorrent-read-file pathname))
	 (info (gethash "info" dict)))
    (format stream "Name: ~A~%" (gethash "name" info))
    (format stream "Created~@[ by ~A~]~@[ on ~A~]~%"
	    (gethash "created by" dict)
	    (let ((epoch (gethash "creation date" dict)))
	      (when epoch
		(multiple-value-bind  (sec min hour day month year wd st tz)
		    (decode-universal-time
		     (unix->universal-time epoch))
		  (declare (ignore wd st tz))
		  (format nil "~A/~A/~A ~A:~A:~A" day month year hour min sec)))))
    (format stream "Encoding: ~A~%" (gethash "encoding" dict))
    (format stream "Tracker: ~A~%" (gethash "announce" dict))
    (format stream "Announce List:~%~{  ~A~%~}"
	    (mapcar #'(lambda (entry)
			(if (consp entry)
			    (car entry)
			    entry))
		    (gethash "announce-list" dict)))
    (let ((total 0)
	  (files (gethash "files" info))
	  (prefix (gethash "name" info)))
      (cond (files
	     (format stream "~D Files:~%" (length files))
	     (dolist (f files)
	       (let ((l (gethash "length" f)))
		 (incf total l)
		 (format stream "  ~10D~14T~@[~A/~]~{~A~^/~}~%"
			 l prefix (gethash "path" f)))))
	    (t
	     (format stream "File: ~A~%" prefix)
	     (setf total (gethash "length" info))))
      (format stream "Total: ~D bytes~%" total)))
  (values))

(defun bittorrent-add-trackers (pathname add-list &key prepend (output pathname))
  (let* ((dict (bittorrent-read-file pathname))
	 (trackers (gethash "announce-list" dict)))
    (dolist (trk add-list)
      (if (assoc trk trackers :test #'string=)
	  (warn "~A already in the announce list." trk)
	  (if prepend
	      (push (list trk)
		    (gethash "announce-list" dict))
	      (setf (gethash "announce-list" dict)
		    (append trackers
			    (list trk))))))
    (bittorrent-write-file output dict :if-exists :supersede))
  (values))
